from __future__ import annotations

from hashlib import md5
from importlib.resources import files
from tempfile import NamedTemporaryFile
from urllib.parse import urlparse
from urllib.request import urlopen

from pandas import DataFrame, Series
from pvlib.iotools import read_epw

from .base import HOURS_DF, SimulationError, missing_required_inputs
from .reporting import find_hour_peak, output_precision, to_output_csv, find_day_peak
from .solar_irradiation import SolarIrradiationSimulation
from ..schemas import LocationSimulationOutput
from ..types import OpenBESSpecification


def get_available_epw_files() -> list[str]:
    climate_data_dir = files("openbes.simulations.climate_data")
    return [f"openbes://{f.name}" for f in climate_data_dir.iterdir() if f.name.endswith(".epw")]


class LocationSimulationError(SimulationError):
    """Raised when location report generation fails."""


class LocationSimulation:
    """Loads EPW data and owns EPW-derived weather/solar properties."""
    _required_inputs = ("meteorological_file_path",)

    def __init__(self, spec: OpenBESSpecification):
        self.spec = spec
        missing = missing_required_inputs(self.spec, self._required_inputs)
        if missing:
            raise LocationSimulationError(
                f"LocationSimulation missing required inputs: {', '.join(missing)}"
            )
        self._source_path: str | None = None
        self._epw_data: DataFrame | None = None
        self._epw_metadata: dict | None = None
        self._epw_file_checksum: str | None = None
        self._solar_irradiation: SolarIrradiationSimulation | None = None

    @property
    def meteorological_file_path(self) -> str:
        return self.spec.meteorological_file_path

    def _ensure_loaded(self) -> None:
        source = self.meteorological_file_path
        if self._source_path == source and self._epw_data is not None:
            return

        self._source_path = source
        self._epw_data = None
        self._epw_metadata = None
        self._epw_file_checksum = None
        self._solar_irradiation = None

        parsed = urlparse(source)
        if parsed.scheme in ("http", "https", "ftp"):
            # Try to load pyodide.http packages to check if we're in a Pyodide environment.
            try:
                with urlopen(source) as response:
                    content = response.read()
            except Exception as exc:
                raise type(exc)(
                    f"{exc}. If remote EPW access issues persist, download the file locally "
                    "and supply a local file path instead."
                ) from exc
            with NamedTemporaryFile(suffix=".epw") as tmp:
                tmp.write(content)
                tmp.flush()
                self._epw_data, self._epw_metadata = read_epw(tmp.name)
            self._epw_file_checksum = md5(content).hexdigest()
            return

        if source.startswith("openbes://"):
            package_path = source[len("openbes://"):]
            epw_path = files("openbes.simulations.climate_data") / package_path
            content = epw_path.read_bytes()
            self._epw_data, self._epw_metadata = read_epw(str(epw_path))
            self._epw_file_checksum = md5(content).hexdigest()
            return

        # Local file access is allowed
        try:
            self._epw_data, self._epw_metadata = read_epw(source)
            with open(source, "rb") as f:
                content = f.read()
                self._epw_file_checksum = md5(content).hexdigest()
        except Exception as exc:
            raise type(exc)(
                f"Could not load {source} as a local EPW file. Ensure the path is correct and accessible."
            ) from exc

    @property
    def epw_data(self) -> DataFrame:
        self._ensure_loaded()
        return self._epw_data

    @property
    def epw_metadata(self) -> dict:
        self._ensure_loaded()
        return self._epw_metadata

    @property
    def epw_file_checksum(self) -> str:
        self._ensure_loaded()
        return self._epw_file_checksum

    @property
    def dry_bulb_temp(self) -> Series:
        return self.epw_data["temp_air"]

    @property
    def wind_speed(self) -> Series:
        return self.epw_data["wind_speed"]

    @property
    def supply_air_temp(self) -> Series:
        return self.epw_data["temp_air"]

    @property
    def relative_humidity(self) -> Series:
        return self.epw_data["relative_humidity"]

    @property
    def solar_irradiation(self) -> SolarIrradiationSimulation:
        if self._solar_irradiation is None:
            self._solar_irradiation = SolarIrradiationSimulation(
                epw_data=self.epw_data,
                epw_metadata=self.epw_metadata,
            )
        return self._solar_irradiation

    @property
    def _hourly_dry_bulb_temp(self) -> Series:
        return Series(self.dry_bulb_temp.values, index=HOURS_DF.index)

    @property
    def _hourly_ghi(self) -> Series:
        return Series(
            self.solar_irradiation.ghi.values,
            index=HOURS_DF.index,
            name="global_horizontal_irradiance",
        )

    @property
    def report(self) -> LocationSimulationOutput:
        try:
            precision = output_precision(self.spec)
            quantiles = [
                0,
                0.004,
                0.01,
                0.02,
                0.1,
                0.2,
                0.3,
                0.4,
                0.5,
                0.6,
                0.7,
                0.8,
                0.9,
                0.95,
                0.99,
                0.996,
                1,
            ]
            base_temperature = 18.0
            max_daily_t = self._hourly_dry_bulb_temp.groupby(["month", "day"]).max()
            min_daily_t = self._hourly_dry_bulb_temp.groupby(["month", "day"]).min()
            avg_daily_t = (max_daily_t + min_daily_t) / 2
            heating_dd = base_temperature - avg_daily_t
            degree_days = DataFrame(
                {
                    "Heating Degree Days": heating_dd.clip(lower=0),
                    "Cooling Degree Days": (-heating_dd).clip(lower=0),
                },
                index=avg_daily_t.index,
            )
            solstice_mask = (
                self._hourly_ghi.index.get_level_values("month").isin([6, 12])
            ) & (self._hourly_ghi.index.get_level_values("day").isin([172, 355]))
            ghi = self._hourly_ghi.loc[solstice_mask].reset_index()
            ghi["month"] = ghi["month"].map({6: "June 21", 12: "December 21"})
            ghi = ghi.drop(columns="day").pivot(
                index="hour", columns="month", values="global_horizontal_irradiance"
            )
            annual_radiation = (
                (self.solar_irradiation.solar_irradiation.sum(axis="rows") / 1000)
                .rename(index=lambda x: x.value)
                .rename_axis("Compass point")
                .to_frame(name="Annual incident solar radiation (kWh/m2)")
            )
            quantiles_df = (
                self._hourly_dry_bulb_temp.quantile(quantiles)
                .rename_axis("Quantile")
                .to_frame(name="Temperature (C)")
            )
            return LocationSimulationOutput(
                altitude=self.epw_metadata["altitude"],
                latitude=self.epw_metadata["latitude"],
                longitude=self.epw_metadata["longitude"],
                city=self.epw_metadata["city"],
                country=self.epw_metadata["country"],
                state_province=self.epw_metadata["state-prov"],
                solstice_ghr_csv=to_output_csv(ghi, precision),
                max_outdoor_temperature=find_hour_peak(
                    self._hourly_dry_bulb_temp, max, precision
                ),
                min_outdoor_temperature=find_hour_peak(
                    self._hourly_dry_bulb_temp, min, precision
                ),
                mean_outdoor_temperature=round(self._hourly_dry_bulb_temp.mean(), precision),
                max_outdoor_day_temperature=find_day_peak(
                    avg_daily_t, max, precision
                ),
                min_outdoor_day_temperature=find_day_peak(
                    avg_daily_t, min, precision
                ),
                mean_outdoor_day_temperature=round(
                    avg_daily_t.groupby(["month", "day"]).mean().mean(), precision
                ),
                climate_quantiles_csv=to_output_csv(quantiles_df, precision),
                degree_days_csv=to_output_csv(degree_days, precision),
                annual_incident_solar_radiation_csv=to_output_csv(
                    annual_radiation, precision
                ),
            )
        except Exception as exc:
            raise LocationSimulationError("Failed to generate location report") from exc



"""Data classes for OpenBES simulation specifications and parameters.

This will be removed in v2 in favour of the schemas-driven approach.
"""

from dataclasses import dataclass, field
from pathlib import Path
from urllib.parse import urlparse
from typing import Optional, Union
import tomllib

from . import LIGHTING_CONTROL, COOLING_SYSTEM_TYPES, HEATING_SYSTEM_TYPES, ListableEnum
from .enums import (
    LIGHTING_TECHNOLOGIES,
    LIGHTING_BALLASTS,
    ENERGY_SOURCES,
    HEAT_CAPACITY_CLASSES,
    TERRAINS,
)


@dataclass
class OpenBESParameters:
    """
    Data class for OpenBES simulation parameters.
    """

    advanced_heat_capacity_am: Optional[float] = None
    air_heat_capacity: Optional[float] = None
    altitude: Optional[float] = None
    appliance_on_off: Optional[bool] = True
    cooling_load_factor: Optional[float] = 1.0
    cooling_system1_min_demand: Optional[float] = 15.0
    cooling_system2_energy_efficifiency_ratio: Optional[float] = None
    cooling_system2_energy_source: Optional[ENERGY_SOURCES] = field(
        default=None, metadata={"cls": ENERGY_SOURCES}
    )
    cooling_system2_min_demand: Optional[float] = None
    cooling_system2_nominal_capacity: Optional[float] = None
    cooling_system2_number: Optional[int] = 0
    cooling_system2_off_time: Optional[int] = None
    cooling_system2_on_time: Optional[int] = None
    cooling_system2_sensible_nominal_capacity: Optional[float] = None
    cooling_system2_simultaneity_factor_canteen: Optional[float] = 0.0
    cooling_system2_simultaneity_factor_common: Optional[float] = 0.0
    cooling_system2_simultaneity_factor_office: Optional[float] = 0.0
    cooling_system2_simultaneity_factor_other: Optional[float] = 0.0
    cooling_system2_simultaneity_factor_teaching: Optional[float] = 0.0
    cooling_system2_type: Optional[COOLING_SYSTEM_TYPES] = field(
        default=None, metadata={"cls": COOLING_SYSTEM_TYPES}
    )
    courtyard_length: Optional[float] = None
    courtyard_number: Optional[int] = 0
    courtyard_width: Optional[float] = None
    density_of_air: Optional[float] = 1.211
    facade_absorption_coefficient: Optional[float] = 0.6
    facade_correction_factor: Optional[float] = 1.0
    facade_emissivity: Optional[float] = 0.9
    floor_correction_factor: Optional[float] = 1.0
    heat_capacity_correction_factor: Optional[float] = 1.0
    heat_capacity_joule: Optional[float] = None
    heating_load_factor: Optional[float] = 1.0
    heating_system1_min_demand: Optional[float] = 37
    heating_system2_efficiency_cop: Optional[float] = None
    heating_system2_energy_source: Optional[ENERGY_SOURCES] = field(
        default=None, metadata={"cls": ENERGY_SOURCES}
    )
    heating_system2_min_demand: Optional[float] = None
    heating_system2_nominal_capacity: Optional[float] = None
    heating_system2_number: Optional[int] = 0
    heating_system2_off_time: Optional[float] = None
    heating_system2_on_time: Optional[float] = None
    heating_system2_simultaneity_factor_canteen: Optional[float] = 0.0
    heating_system2_simultaneity_factor_common: Optional[float] = 0.0
    heating_system2_simultaneity_factor_office: Optional[float] = 0.0
    heating_system2_simultaneity_factor_other: Optional[float] = 0.0
    heating_system2_simultaneity_factor_teaching: Optional[float] = 0.0
    heating_system2_type: Optional[HEATING_SYSTEM_TYPES] = field(
        default=None, metadata={"cls": HEATING_SYSTEM_TYPES}
    )
    infiltration_correction_factor: Optional[float] = 1.0
    leakage_air_flow_dependent: Optional[float] = 0.0001
    lighting_on_off: Optional[bool] = True
    # Fraction of gross building area that is inhabitable (i.e. minus walls, shafts, etc)
    nia_gba_ratio: Optional[float] = 0.95
    occupancy_close_common: Optional[float] = None
    occupancy_close_other: Optional[float] = None
    occupancy_on_off: Optional[bool] = True
    occupancy_open_common: Optional[float] = None
    occupancy_open_other: Optional[float] = None
    open_courtyard_depth_a1: Optional[float] = None
    open_courtyard_depth_b1: Optional[float] = None
    open_courtyard_depth_c1: Optional[float] = None
    open_courtyard_depth_d1: Optional[float] = None
    open_courtyard_number_a1: Optional[int] = 0
    open_courtyard_number_b1: Optional[int] = 0
    open_courtyard_number_c1: Optional[int] = 0
    open_courtyard_number_d1: Optional[int] = 0
    output_csv_precision: Optional[int] = 2
    pressure_of_air: Optional[float] = None
    roof_absorption_coefficient: Optional[float] = 0.8
    roof_correction_factor: Optional[float] = 1.0
    roof_emissivity: Optional[float] = 0.9
    shading_correction_factor: Optional[float] = 1.0
    specific_heat_of_air: Optional[float] = 1.015
    ventilation_system2_airflow: Optional[float] = None
    ventilation_system2_energy_source: Optional[float] = None
    ventilation_system2_heat_recovery_efficiency: Optional[float] = None
    ventilation_system2_off_time: Optional[float] = None
    ventilation_system2_on_time: Optional[float] = None
    ventilation_system2_rated_input_power: Optional[float] = None
    ventilation_system2_type: Optional[float] = None
    ventilation_system2_ventilated_area: Optional[float] = None
    view_factor_to_sky_facade: Optional[float] = 0.5
    view_factor_to_sky_roof: Optional[float] = 1.0
    window_correction_factor: Optional[float] = 1.0
    window_optical_c1: Optional[float] = 1.0
    window_optical_c2: Optional[float] = -0.189
    window_optical_c3: Optional[float] = 0.644
    window_optical_c4: Optional[float] = -0.596
    window_optical_c5: Optional[float] = 0.0

    def __post_init__(self):
        super().__init__()
        for f in self.__dataclass_fields__.values():
            value = getattr(self, f.name)
            cls = f.metadata.get("cls", None)
            if isinstance(cls, ListableEnum) and not isinstance(value, cls):
                if isinstance(value, str):
                    setattr(self, f.name, cls.get_by_value(value))


def get_meteorological_file(filename: str) -> str:
    if "Denver" in filename:
        return "USA_Denver_725650TYCST.epw"
    if "Oxford" in filename:
        return "UK_Oxford_GBR_ENG_RAF.Benson.036580_TMYx.2007-2021.epw"
    if "Sevilla" in filename:
        return "SPAIN_Sevilla.083910_SWEC.epw"
    if "Madrid" in filename:
        return "SPAIN_Madrid.082210_SWEC.epw"
    raise ValueError(f"Unknown meteorological file: {filename}")


def get_meteorological_name(epw_filename: str) -> str:
    if "Denver" in epw_filename:
        return "725650_Denver"
    if "Oxford" in epw_filename:
        return "Oxford"
    if "Sevilla" in epw_filename:
        return "Sevilla"
    if "Madrid" in epw_filename:
        return "Madrid"
    return epw_filename


def normalize_meteorological_file_path(path: str) -> str:
    """Normalize legacy and modern meteorological path values to a canonical path string."""
    if path is None:
        return path
    p = str(path).strip()
    if not p:
        return p

    parsed = urlparse(p)
    has_scheme = bool(parsed.scheme)
    has_path_sep = "/" in p or "\\" in p

    if has_scheme:
        return p

    # Legacy short names only apply to bare identifiers.
    if not has_path_sep:
        try:
            p = get_meteorological_file(p)
        except ValueError:
            pass

    # Raw filenames should resolve to package data by default.
    if not has_path_sep:
        return f"openbes://{p}"
    return p


def meteorological_file_path_to_toml_value(path: str) -> str:
    """Convert canonical path to legacy TOML value conventions."""
    if not path:
        return ""
    if path.startswith("openbes://"):
        package_path = path[len("openbes://"):]
        if "/" not in package_path and "\\" not in package_path:
            return package_path
    return path


@dataclass
class OpenBESSpecification:
    """
    Data class for OpenBES spec parameters.
    """

    building_length: Optional[float] = None
    building_width: Optional[float] = None
    meteorological_file_path: Optional[str] = None

    parameters: Optional[OpenBESParameters] = field(default_factory=OpenBESParameters)
    appliances_load: float = 0.0
    biomass_annual: Optional[float] = None
    biomass_pellets_annual: Optional[float] = None
    building_area: Optional[str] = None
    building_height: Optional[float] = None
    building_name: Optional[str] = None
    building_standby_load: Optional[float] = 0.0
    building_type: Optional[str] = None
    # Conditioned=True, Unconditioned=False
    condition_z1: Optional[bool] = True
    # Conditioned=True, Unconditioned=False
    condition_z2: Optional[bool] = True
    # Conditioned=True, Unconditioned=False
    condition_z3: Optional[bool] = True
    # Conditioned=True, Unconditioned=False
    condition_z4: Optional[bool] = None
    # Conditioned=True, Unconditioned=False
    condition_z5: Optional[bool] = None
    cooling_system1_energy_efficifiency_ratio: Optional[float] = None
    cooling_system1_energy_source: Optional[ENERGY_SOURCES] = field(
        default=None, metadata={"cls": ENERGY_SOURCES}
    )
    cooling_system1_nominal_capacity: Optional[float] = None
    cooling_system1_number: Optional[int] = 0
    cooling_system1_off_time: Optional[int] = None
    cooling_system1_on_time: Optional[int] = None
    cooling_system1_sensible_nominal_capacity: Optional[float] = None
    cooling_system1_simultaneity_factor_canteen: Optional[float] = 0.0
    cooling_system1_simultaneity_factor_common: Optional[float] = 0.0
    cooling_system1_simultaneity_factor_office: Optional[float] = 0.0
    cooling_system1_simultaneity_factor_other: Optional[float] = 0.0
    cooling_system1_simultaneity_factor_teaching: Optional[float] = 0.0
    cooling_system1_type: Optional[COOLING_SYSTEM_TYPES] = field(
        default=None, metadata={"cls": COOLING_SYSTEM_TYPES}
    )
    country: Optional[str] = None
    diesel_annual: Optional[float] = None
    electricity_annual: Optional[float] = None
    electricity_april: Optional[float] = None
    electricity_august: Optional[float] = None
    electricity_december: Optional[float] = None
    electricity_february: Optional[float] = None
    electricity_january: Optional[float] = None
    electricity_july: Optional[float] = None
    electricity_june: Optional[float] = None
    electricity_march: Optional[float] = None
    electricity_may: Optional[float] = None
    electricity_november: Optional[float] = None
    electricity_october: Optional[float] = None
    electricity_september: Optional[float] = None
    energy_generated: Optional[float] = None
    energy_used: Optional[float] = None
    first_floor_area_z1: Optional[float] = None
    first_floor_area_z2: Optional[float] = None
    first_floor_area_z3: Optional[float] = None
    first_floor_area_z4: Optional[float] = None
    first_floor_area_z5: Optional[float] = None
    floor_to_ceiling_height: float = 2.7
    fourth_floor_area_z1: Optional[float] = None
    fourth_floor_area_z2: Optional[float] = None
    fourth_floor_area_z3: Optional[float] = None
    fourth_floor_area_z4: Optional[float] = None
    fourth_floor_area_z5: Optional[float] = None
    gas_april: Optional[float] = None
    gas_august: Optional[float] = None
    gas_december: Optional[float] = None
    gas_february: Optional[float] = None
    gas_january: Optional[float] = None
    gas_july: Optional[float] = None
    gas_june: Optional[float] = None
    gas_march: Optional[float] = None
    gas_may: Optional[float] = None
    gas_november: Optional[float] = None
    gas_october: Optional[float] = None
    gas_september: Optional[float] = None
    ground_floor_area_z1: Optional[float] = None
    ground_floor_area_z2: Optional[float] = None
    ground_floor_area_z3: Optional[float] = None
    ground_floor_area_z4: Optional[float] = None
    ground_floor_area_z5: Optional[float] = None
    heat_capacity: Optional[HEAT_CAPACITY_CLASSES] = field(
        default=None, metadata={"cls": HEAT_CAPACITY_CLASSES}
    )
    heating_system1_efficiency_cop: Optional[float] = None
    heating_system1_energy_source: Optional[ENERGY_SOURCES] = field(
        default=None, metadata={"cls": ENERGY_SOURCES}
    )
    heating_system1_nominal_capacity: Optional[float] = None
    heating_system1_number: Optional[int] = 0
    heating_system1_off_time: Optional[float] = None
    heating_system1_on_time: Optional[float] = None
    heating_system1_simultaneity_factor_canteen: Optional[float] = 0.0
    heating_system1_simultaneity_factor_common: Optional[float] = 0.0
    heating_system1_simultaneity_factor_office: Optional[float] = 0.0
    heating_system1_simultaneity_factor_other: Optional[float] = 0.0
    heating_system1_simultaneity_factor_teaching: Optional[float] = 0.0
    heating_system1_type: Optional[HEATING_SYSTEM_TYPES] = field(
        default=None, metadata={"cls": HEATING_SYSTEM_TYPES}
    )
    holiday: Optional[bool] = None
    leakage_air_flow: Optional[float] = None
    leakage_air_flow_independent: float = 0.0
    lighting_control: Optional[LIGHTING_CONTROL] = field(
        default=None, metadata={"cls": LIGHTING_CONTROL}
    )
    lighting_off_time: Optional[float] = None
    lighting_on_time: Optional[float] = None
    lighting_simultaneity_factor: Optional[float] = 0.2
    lighting_system_ballast_z1: Optional[LIGHTING_BALLASTS] = field(
        default=None, metadata={"cls": LIGHTING_BALLASTS}
    )
    lighting_system_ballast_z2: Optional[LIGHTING_BALLASTS] = field(
        default=None, metadata={"cls": LIGHTING_BALLASTS}
    )
    lighting_system_ballast_z3: Optional[LIGHTING_BALLASTS] = field(
        default=None, metadata={"cls": LIGHTING_BALLASTS}
    )
    lighting_system_ballast_z4: Optional[LIGHTING_BALLASTS] = field(
        default=None, metadata={"cls": LIGHTING_BALLASTS}
    )
    lighting_system_ballast_z5: Optional[LIGHTING_BALLASTS] = field(
        default=None, metadata={"cls": LIGHTING_BALLASTS}
    )
    lighting_system_ballast_z6: Optional[LIGHTING_BALLASTS] = field(
        default=None, metadata={"cls": LIGHTING_BALLASTS}
    )
    lighting_system_lamp_number_z1: Optional[int] = None
    lighting_system_lamp_number_z2: Optional[int] = None
    lighting_system_lamp_number_z3: Optional[int] = None
    lighting_system_lamp_number_z4: Optional[int] = None
    lighting_system_lamp_number_z5: Optional[int] = None
    lighting_system_lamp_number_z6: Optional[int] = None
    lighting_system_lamp_power_z1: Optional[float] = None
    lighting_system_lamp_power_z2: Optional[float] = None
    lighting_system_lamp_power_z3: Optional[float] = None
    lighting_system_lamp_power_z4: Optional[float] = None
    lighting_system_lamp_power_z5: Optional[float] = None
    lighting_system_lamp_power_z6: Optional[float] = None
    lighting_system_luminary_number_z1: Optional[int] = None
    lighting_system_luminary_number_z2: Optional[int] = None
    lighting_system_luminary_number_z3: Optional[int] = None
    lighting_system_luminary_number_z4: Optional[int] = None
    lighting_system_luminary_number_z5: Optional[int] = None
    lighting_system_luminary_number_z6: Optional[int] = None
    lighting_system_name_z1: Optional[str] = None
    lighting_system_name_z2: Optional[str] = None
    lighting_system_name_z3: Optional[str] = None
    lighting_system_name_z4: Optional[str] = None
    lighting_system_name_z5: Optional[str] = None
    lighting_system_name_z6: Optional[str] = None
    lighting_system_operating_hours_z1: Optional[float] = None
    lighting_system_operating_hours_z2: Optional[float] = None
    lighting_system_operating_hours_z3: Optional[float] = None
    lighting_system_operating_hours_z4: Optional[float] = None
    lighting_system_operating_hours_z5: Optional[float] = None
    lighting_system_operating_hours_z6: Optional[float] = None
    lighting_system_similar_zone_number_z1: Optional[int] = None
    lighting_system_similar_zone_number_z2: Optional[int] = None
    lighting_system_similar_zone_number_z3: Optional[int] = None
    lighting_system_similar_zone_number_z4: Optional[int] = None
    lighting_system_similar_zone_number_z5: Optional[int] = None
    lighting_system_similar_zone_number_z6: Optional[int] = None
    lighting_system_simultaneity_factor_z1: Optional[float] = 0.0
    lighting_system_simultaneity_factor_z2: Optional[float] = 0.0
    lighting_system_simultaneity_factor_z3: Optional[float] = 0.0
    lighting_system_simultaneity_factor_z4: Optional[float] = 0.0
    lighting_system_simultaneity_factor_z5: Optional[float] = 0.0
    lighting_system_simultaneity_factor_z6: Optional[float] = 0.0
    lighting_system_tech_z1: Optional[LIGHTING_TECHNOLOGIES] = field(
        default=None, metadata={"cls": LIGHTING_TECHNOLOGIES}
    )
    lighting_system_tech_z2: Optional[LIGHTING_TECHNOLOGIES] = field(
        default=None, metadata={"cls": LIGHTING_TECHNOLOGIES}
    )
    lighting_system_tech_z3: Optional[LIGHTING_TECHNOLOGIES] = field(
        default=None, metadata={"cls": LIGHTING_TECHNOLOGIES}
    )
    lighting_system_tech_z4: Optional[LIGHTING_TECHNOLOGIES] = field(
        default=None, metadata={"cls": LIGHTING_TECHNOLOGIES}
    )
    lighting_system_tech_z5: Optional[LIGHTING_TECHNOLOGIES] = field(
        default=None, metadata={"cls": LIGHTING_TECHNOLOGIES}
    )
    lighting_system_tech_z6: Optional[LIGHTING_TECHNOLOGIES] = field(
        default=None, metadata={"cls": LIGHTING_TECHNOLOGIES}
    )
    location: Optional[float] = None
    LPG_annual: Optional[float] = None
    max_building_occupation: Optional[float] = None
    natural_gas_annual: Optional[float] = None
    natural_ventilation_night: Optional[float] = 0.0
    occupancy_close_canteen: Optional[float] = None
    occupancy_close_office: Optional[float] = None
    occupancy_close_teaching: Optional[float] = None
    occupancy_open_canteen: Optional[float] = None
    occupancy_open_office: Optional[float] = None
    occupancy_open_teaching: Optional[float] = None
    orientation_angle: float = 0.0
    other_electricity_usage: Optional[float] = 0.0
    other_gas_usage: Optional[float] = 0.0
    roof_angle: float = 0.0
    schedule_april: Optional[bool] = None
    schedule_august: Optional[bool] = None
    schedule_december: Optional[bool] = None
    schedule_february: Optional[bool] = None
    schedule_friday: Optional[bool] = None
    schedule_january: Optional[bool] = None
    schedule_july: Optional[bool] = None
    schedule_june: Optional[bool] = None
    schedule_march: Optional[bool] = None
    schedule_may: Optional[bool] = None
    schedule_monday: Optional[bool] = None
    schedule_november: Optional[bool] = None
    schedule_october: Optional[bool] = None
    schedule_saturday: Optional[bool] = None
    schedule_september: Optional[bool] = None
    schedule_sunday: Optional[bool] = None
    schedule_thursday: Optional[bool] = None
    schedule_tuesday: Optional[bool] = None
    schedule_wednesday: Optional[bool] = None
    second_floor_area_z1: Optional[float] = None
    second_floor_area_z2: Optional[float] = None
    second_floor_area_z3: Optional[float] = None
    second_floor_area_z4: Optional[float] = None
    second_floor_area_z5: Optional[float] = None
    # Better described as 'setpoint maximum'
    setpoint_summer_day: Optional[float] = 21.0
    # Better described as 'setpoint maximum'
    setpoint_summer_night: Optional[float] = 21.0
    # Better described as 'setpoint minimum'
    setpoint_winter_day: Optional[float] = 21.0
    # Better described as 'setpoint minimum'
    setpoint_winter_night: Optional[float] = 21.0
    slab_thickness: float = 0.2
    solar_external_shading_summer: float = 0.0
    solar_external_shading_winter: float = 0.0
    terrain_class: Optional[TERRAINS] = field(default=None, metadata={"cls": TERRAINS})
    thermal_bridge_facade_ground: Optional[bool] = None
    thermal_bridge_facade_intermediate: Optional[bool] = None
    thermal_bridge_facade_roof: Optional[bool] = None
    thermal_bridge_shading: Optional[bool] = None
    thermal_bridge_window: Optional[bool] = None
    third_floor_area_z1: Optional[float] = None
    third_floor_area_z2: Optional[float] = None
    third_floor_area_z3: Optional[float] = None
    third_floor_area_z4: Optional[float] = None
    third_floor_area_z5: Optional[float] = None
    typical_occupation: Optional[float] = None
    uvalue_facade: float = 0.0
    uvalue_floor: float = 0.0
    uvalue_roof: float = 0.0
    uvalue_window: float = 0.0
    ventilation_system1_airflow: Optional[float] = 0.0
    ventilation_system1_energy_source: Optional[ENERGY_SOURCES] = field(
        default=None, metadata={"cls": ENERGY_SOURCES}
    )
    ventilation_system1_heat_recovery_efficiency: Optional[float] = 0.0
    ventilation_system1_off_time: Optional[int] = 1
    ventilation_system1_on_time: Optional[int] = 24
    ventilation_system1_rated_input_power: Optional[float] = 0.0
    ventilation_system1_type: Optional[float] = None
    ventilation_system1_ventilated_area: Optional[float] = 1.0
    water_demand: Optional[float] = None
    water_reference_temperature: Optional[float] = None
    water_supply_temperature: Optional[float] = None
    water_system_efficiency_cop: Optional[float] = None
    water_system_energy_source: Optional[ENERGY_SOURCES] = field(
        default=None, metadata={"cls": ENERGY_SOURCES}
    )
    water_system_nominal_capacity: Optional[float] = None
    water_system_type: Optional[float] = None
    window_frame_factor: float = 0.0
    window_gvalue: float = 0.0
    window_height: float = 0.0
    window_length: float = 0.0
    window_number_first_a1: Optional[int] = 0
    window_number_first_b1: Optional[int] = 0
    window_number_first_c1: Optional[int] = 0
    window_number_first_d1: Optional[int] = 0
    window_number_fourth_a1: Optional[int] = 0
    window_number_fourth_b1: Optional[int] = 0
    window_number_fourth_c1: Optional[int] = 0
    window_number_fourth_d1: Optional[int] = 0
    window_number_ground_a1: Optional[int] = 0
    window_number_ground_b1: Optional[int] = 0
    window_number_ground_c1: Optional[int] = 0
    window_number_ground_d1: Optional[int] = 0
    window_number_second_a1: Optional[int] = 0
    window_number_second_b1: Optional[int] = 0
    window_number_second_c1: Optional[int] = 0
    window_number_second_d1: Optional[int] = 0
    window_number_third_a1: Optional[int] = 0
    window_number_third_b1: Optional[int] = 0
    window_number_third_c1: Optional[int] = 0
    window_number_third_d1: Optional[int] = 0
    zone_name_z1: Optional[float] = None
    zone_name_z2: Optional[float] = None
    zone_name_z3: Optional[float] = None
    zone_name_z4: Optional[float] = None
    zone_name_z5: Optional[float] = None

    def __post_init__(self):
        super().__init__()
        self.meteorological_file_path = normalize_meteorological_file_path(
            self.meteorological_file_path
        )
        if self.parameters is not None and not isinstance(
            self.parameters, OpenBESParameters
        ):
            self.parameters = OpenBESParameters(**self.parameters)
        for f in self.__dataclass_fields__.values():
            value = getattr(self, f.name)
            cls = f.metadata.get("cls", None)
            if cls and issubclass(cls, ListableEnum) and not isinstance(value, cls):
                if isinstance(value, str):
                    setattr(self, f.name, cls.get_by_value(value))

    @classmethod
    def from_toml(cls, toml_file: Union[dict, str, Path]):
        if isinstance(toml_file, (str, Path)):
            with open(toml_file, "rb") as f:
                toml_content = tomllib.load(f)
        else:
            toml_content = toml_file

        filtered = {k: v for k, v in toml_content.items() if v is not None and v != ""}
        typed = dict(filtered)
        for k, v in typed.items():
            if k == "i.meteorological_file":
                typed[k] = normalize_meteorological_file_path(v)
            elif isinstance(v, str) and v.lower() in ["false", "no"]:
                typed[k] = False
            elif isinstance(v, str) and v.lower() in ["true", "yes"]:
                typed[k] = True
            elif k.startswith(("i.condition_z", "d.condition_z")):
                typed[k] = v.lower() == "conditioned"

        parameters = {k[2:]: v for k, v in typed.items() if k.startswith("d")}
        specification = {k[2:]: v for k, v in typed.items() if k.startswith("i")}
        if "meteorological_file" in specification:
            specification["meteorological_file_path"] = specification.pop("meteorological_file")
        return cls(parameters=OpenBESParameters(**parameters), **specification)



"""Converting legacy TOML to JSON.

The historical TOML format used by OpenBES only captures a flattened view of
the data model.  The new JSON Schema-based data
model is richer (allowing arbitrary list lengths, nested structures, etc.), so
these helpers provide a best-effort conversion layer.
"""

from __future__ import annotations

import tomllib
from pathlib import Path
from typing import Any, Dict, Optional, List

from openbes.schemas import OpenBESSpecificationV2
from openbes.types import normalize_meteorological_file_path


def monthly_average_to_consumption(value: float = None) -> Optional[Dict[str, float]]:
    if value is None:
        return None
    return {
        "January": value,
        "February": value,
        "March": value,
        "April": value,
        "May": value,
        "June": value,
        "July": value,
        "August": value,
        "September": value,
        "October": value,
        "November": value,
        "December": value,
    }


def annual_to_consumption(value: float) -> Optional[Dict[str, float]]:
    if value is None:
        return None
    return {k: v / 12.0 for k, v in monthly_average_to_consumption(value).items()}


def toml_to_json(toml: dict | Path | str, allow_warnings: bool = True) -> dict:
    """Convert a TOML dictionary to an OpenBES.schema.json dictionary.

    Only keys that are actually present in the TOML input are emitted; no
    synthetic defaults are introduced.  An empty TOML input ({}) therefore
    produces an empty JSON output ({}).

    Args:
        toml: The input TOML dictionary, file path, or string content.
        allow_warnings: If False, any warnings encountered during conversion will raise an exception.

    Returns:
        A dictionary suitable for constructing an OpenBESSpecificationV2. The json.dump() of the dictionary
        will match be valid against the OpenBES.schema.json specification.

    Raises:
        ValueError: If allow_warnings is False and any warnings were encountered during conversion.
        ValidationError: If the resulting dictionary does not validate against OpenBESSpecificationV2.
    """
    warnings: list[str] = []

    content = toml
    if isinstance(toml, (Path, str)):
        if Path(toml).is_file():
            with open(toml, "r") as f:
                content = tomllib.loads(f.read())
        else:
            content = tomllib.loads(toml)

    # Helper functions
    def get(key: str, default: Any = None) -> Any:
        """Helper to get a value from a TOML dictionary with heuristic coercion.
        Returns *default* (None by default) when the key is absent or blank.
        """
        if f"i.{key}" in content:
            value = content[f"i.{key}"]
        elif f"d.{key}" in content:
            value = content[f"d.{key}"]
        else:
            return default
        if value == "" or value is None:
            return default
        if isinstance(value, str):
            vs = value.strip()
            if vs.lower() in ("yes", "y", "true", "t", "1"):
                return True
            if vs.lower() in ("no", "n", "false", "f", "0"):
                return False
            try:
                if "." in vs:
                    return float(vs)
                return int(vs)
            except ValueError:
                return vs
        return value

    def present(key: str) -> bool:
        """Return True if the key exists in the TOML content with a non-blank value."""
        return get(key) is not None

    def to_range(start_key: str, end_key: str) -> Optional[Dict[str, float]]:
        start = get(start_key)
        end = get(end_key)
        if start is None or end is None:
            if start is not None or end is not None:
                warnings.append(
                    f"Ignoring range because {start_key if start is None else end_key} is not set."
                )
            return None
        return {"min": start, "max": end}

    def to_duration(start_key: str, end_key: str) -> Optional[Dict[str, float]]:
        start = get(start_key)
        end = get(end_key)
        if start is None or end is None:
            if start is not None or end is not None:
                warnings.append(
                    f"Ignoring duration because {start_key if start is None else end_key} is not set."
                )
            return None
        return {"start": start, "end": end}

    def has_required_keys(obj: Any, required_keys: List[str], context: str) -> bool:
        missing_keys = [k for k in required_keys if obj.get(k) is None]
        if len(missing_keys) > 0 and len(missing_keys) != len(required_keys):
            warnings.append(
                f"Skipping incomplete {context} definition.\n"
                f"\tSpecified keys: {', '.join([r for r in required_keys if r not in missing_keys])};\n"
                f"\tMissing missing keys: {', '.join(missing_keys)}."
            )
            return False
        return not any(missing_keys)

    out: dict = {}

    # ── parameters ────────────────────────────────────────────────────────────
    parameters: dict = {}

    for k in [
        "altitude",
        "cooling_load_factor",
        "density_of_air",
        "facade_absorption_coefficient",
        "facade_correction_factor",
        "facade_emissivity",
        "floor_correction_factor",
        "heat_capacity_correction_factor",
        "heat_capacity_joule",
        "heating_load_factor",
        "infiltration_correction_factor",
        "leakage_air_flow_dependent",
        "nia_gba_ratio",
        "pressure_of_air",
        "roof_absorption_coefficient",
        "roof_correction_factor",
        "roof_emissivity",
        "shading_correction_factor",
        "specific_heat_of_air",
        "view_factor_to_sky_facade",
        "view_factor_to_sky_roof",
        "window_correction_factor",
    ]:
        v = get(k)
        if v is not None:
            parameters[k] = v

    for k in ["lighting", "occupancy"]:
        val = get(f"{k}_on_off")
        if val is not None:
            parameters[f"include_{k}"] = val

    appliance_val = get("appliance_on_off")
    if appliance_val is not None:
        parameters["include_appliances"] = appliance_val

    # Only include window_angular_correction_factors when at least one c-value
    # was explicitly set in the TOML.
    wacf_raw = [get(f"window_optical_c{i}") for i in range(1, 6)]
    if any(v is not None for v in wacf_raw):
        parameters["window_angular_correction_factors"] = [v if v is not None else 0.0 for v in wacf_raw]

    if parameters:
        out["parameters"] = parameters

    # ── simple top-level scalars ───────────────────────────────────────────────
    for k in [
        "appliances_load",
        "country",
        "holiday",
        "leakage_air_flow",
        "leakage_air_flow_independent",
        "lighting_control",
        "lighting_simultaneity_factor",
        "max_building_occupation",
        "natural_ventilation_night",
        "typical_occupation",
    ]:
        v = get(k)
        if v is not None:
            out[k] = v

    # ── consumption fields ─────────────────────────────────────────────────────
    for k in [
        "biomass",
        "biomass_pellets",
        "diesel",
        "electricity",
        "LPG",
        "natural_gas",
    ]:
        annual_value = get(f"{k}_annual")
        if annual_value is not None:
            out[f"{k}_consumption"] = annual_to_consumption(annual_value)

    # Monthly electricity overrides annual if any monthly value is present and non-zero
    monthly_electricity = {
        k: get(f"electricity_{k.lower()}")
        for k in monthly_average_to_consumption(0).keys()
    }
    if (
        any(v is not None for v in monthly_electricity.values())
        and sum(v or 0 for v in monthly_electricity.values()) > 0
    ):
        out["electricity_consumption"] = monthly_electricity

    monthly_gas = {
        k: get(f"gas_{k.lower()}") for k in monthly_average_to_consumption(0).keys()
    }
    if (
        any(v is not None for v in monthly_gas.values())
        and sum(v or 0 for v in monthly_gas.values()) > 0
    ):
        out["natural_gas_consumption"] = monthly_gas

    for k in ["other_electricity_usage", "other_gas_usage"]:
        monthly_value = get(k)
        if monthly_value is not None:
            out[f"{k.replace('_usage', '')}_consumption"] = (
                monthly_average_to_consumption(monthly_value)
            )

    building_standby_raw = get("building_standby_load")
    if building_standby_raw is not None:
        out["building_standby_electricity_consumption"] = monthly_average_to_consumption(
            building_standby_raw
        )

    for k in ["energy_generated", "energy_used"]:
        monthly_value = get(k)
        if monthly_value is not None:
            out[k] = monthly_average_to_consumption(monthly_value)

    # ── heat_capacity ──────────────────────────────────────────────────────────
    heat_capacity = get("heat_capacity")
    if heat_capacity == "Very light":
        out["heat_capacity"] = {"Am": 2.5, "Cm": 80_000.0}
    elif heat_capacity == "Light":
        out["heat_capacity"] = {"Am": 2.5, "Cm": 110_000.0}
    elif heat_capacity == "Medium":
        out["heat_capacity"] = {"Am": 2.5, "Cm": 165_000.0}
    elif heat_capacity == "Heavy":
        out["heat_capacity"] = {"Am": 3.0, "Cm": 260_000.0}
    elif heat_capacity == "Very heavy":
        out["heat_capacity"] = {"Am": 3.5, "Cm": 370_000.0}
    elif heat_capacity is not None:
        # Custom numeric value — only emit if we actually have it
        out["heat_capacity"] = {"Am": heat_capacity, "Cm": None}

    # ── lighting active hours ──────────────────────────────────────────────────
    lighting_hours = to_duration("lighting_on_time", "lighting_off_time")
    if lighting_hours is not None:
        out["lighting_active_hours"] = lighting_hours

    # ── meteorological file ────────────────────────────────────────────────────
    mf = get("meteorological_file")
    if mf is not None:
        out["meteorological_file_path"] = normalize_meteorological_file_path(mf)

    # ── setpoint temperatures ──────────────────────────────────────────────────
    for k in ["day", "night"]:
        rng = to_range(f"setpoint_summer_{k}", f"setpoint_winter_{k}")
        if rng is not None:
            out[f"setpoint_temperature_{k}"] = rng

    # ── building ───────────────────────────────────────────────────────────────
    building: dict = {}
    for k in ["area", "height", "length", "width", "name", "type"]:
        v = get(f"building_{k}")
        if v is not None:
            building[k] = v

    for k in [
        "floor_to_ceiling_height",
        "slab_thickness",
        "orientation_angle",
        "roof_angle",
        "solar_external_shading_summer",
        "solar_external_shading_winter",
        "terrain_class",
        "thermal_bridge_facade_ground",
        "thermal_bridge_facade_intermediate",
        "thermal_bridge_facade_roof",
        "thermal_bridge_shading",
        "thermal_bridge_window",
        "uvalue_facade",
        "uvalue_roof",
        "uvalue_window",
        "uvalue_floor",
        "window_frame_factor",
        "window_gvalue",
        "window_height",
        "window_length",
    ]:
        v = get(k)
        if v is not None:
            building[k] = v

    # window_counts: emit a side whenever any of its TOML keys are present
    # (value 0 is valid and must round-trip, so we can't guard on non-zero).
    window_counts: dict = {}
    for o_code, o_name in [
        ("a", "front"),
        ("b", "right"),
        ("c", "back"),
        ("d", "left"),
    ]:
        floors = ["ground", "first", "second", "third", "fourth"]
        if any(present(f"window_number_{f}_{o_code}1") for f in floors):
            window_counts[o_name] = [get(f"window_number_{f}_{o_code}1", 0) for f in floors]
    if window_counts:
        building["window_counts"] = window_counts

    if building:
        out["building"] = building

    # ── zones ─────────────────────────────────────────────────────────────────
    zone_map = ["office", "teaching", "canteen", "common", "other"]
    zones = []
    floor_count = 0
    for i, z in enumerate(zone_map, start=1):
        zone_name = get(f"zone_name_z{i}")
        if zone_name is None:
            # Skip zones with no name set at all
            continue
        zone_areas = [
            get(f"{f}_floor_area_z{i}", 0.0)
            for f in ["ground", "first", "second", "third", "fourth"]
        ]
        floor_count = max(floor_count, sum(1 for a in zone_areas if a is not None and a > 0))
        zone: dict = {
            "name": zone_name,
            "areas": zone_areas,
        }
        condition_val = get(f"condition_z{i}")
        if condition_val is not None:
            zone["conditioned"] = str(condition_val).lower() == "conditioned"
        active = to_duration(f"occupancy_open_{z}", f"occupancy_close_{z}")
        if active is not None:
            zone["active_hours"] = active
        zones.append(zone)

    if zones:
        out["zones"] = zones

    if floor_count > 0:
        out["building"]["floor_count"] = floor_count

    # ── occupation_schedule ────────────────────────────────────────────────────
    schedule_keys = [
        "January", "February", "March", "April", "May", "June",
        "July", "August", "September", "October", "November", "December",
        "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday",
    ]
    occupation_schedule = {}
    for k in schedule_keys:
        v = get(f"schedule_{k.lower()}")
        if v is not None:
            occupation_schedule[k] = bool(v)
    if occupation_schedule:
        out["occupation_schedule"] = occupation_schedule

    # ── heating systems ────────────────────────────────────────────────────────
    heating_systems = []
    for i in [1, 2]:
        system = {
            "type": get(f"heating_system{i}_type"),
            "energy_source": get(f"heating_system{i}_energy_source"),
            "efficiency_cop": get(f"heating_system{i}_efficiency_cop"),
            "min_demand": get(f"heating_system{i}_min_demand"),
            "nominal_capacity": get(f"heating_system{i}_nominal_capacity"),
            "count": get(f"heating_system{i}_number"),
            "active_hours": to_duration(
                f"heating_system{i}_on_time", f"heating_system{i}_off_time"
            ),
            "simultaneity": {
                z["name"]: get(
                    f"heating_system{i}_simultaneity_factor_{zone_map[idx]}", 0.0
                )
                for idx, z in enumerate(zones)
            },
        }
        if has_required_keys(
            system,
            ["energy_source", "efficiency_cop", "nominal_capacity"],
            f"heating system {i}",
        ):
            heating_systems.append(system)
    if heating_systems:
        out["heating_systems"] = heating_systems

    # ── cooling systems ────────────────────────────────────────────────────────
    cooling_systems = []
    for i in [1, 2]:
        system = {
            "type": get(f"cooling_system{i}_type"),
            "energy_source": get(f"cooling_system{i}_energy_source"),
            "efficiency_ratio": get(f"cooling_system{i}_energy_efficifiency_ratio"),
            "min_demand": get(f"cooling_system{i}_min_demand"),
            "nominal_capacity": get(f"cooling_system{i}_nominal_capacity"),
            "sensible_nominal_capacity": get(
                f"cooling_system{i}_sensible_nominal_capacity"
            ),
            "count": get(f"cooling_system{i}_number"),
            "active_hours": to_duration(
                f"cooling_system{i}_on_time", f"cooling_system{i}_off_time"
            ),
            "simultaneity": {
                z["name"]: get(
                    f"cooling_system{i}_simultaneity_factor_{zone_map[idx]}", 0.0
                )
                for idx, z in enumerate(zones)
            },
        }
        if has_required_keys(
            system,
            [
                "energy_source",
                "efficiency_ratio",
                "nominal_capacity",
                "sensible_nominal_capacity",
            ],
            f"cooling system {i}",
        ):
            cooling_systems.append(system)
    if cooling_systems:
        out["cooling_systems"] = cooling_systems

    # ── ventilation systems ────────────────────────────────────────────────────
    ventilation_systems = []
    for i in [1, 2]:
        system = {
            "airflow": get(f"ventilation_system{i}_airflow"),
            "energy_source": get(f"ventilation_system{i}_energy_source"),
            "heat_recovery_efficiency": get(
                f"ventilation_system{i}_heat_recovery_efficiency"
            ),
            "active_hours": to_duration(
                f"ventilation_system{i}_on_time", f"ventilation_system{i}_off_time"
            ),
            "rated_input_power": get(f"ventilation_system{i}_rated_input_power"),
            "type": get(f"ventilation_system{i}_type"),
            "ventilated_area": get(f"ventilation_system{i}_ventilated_area"),
        }
        if has_required_keys(
            system, ["energy_source", "airflow"], f"ventilation system {i}"
        ):
            ventilation_systems.append(system)
    if ventilation_systems:
        out["ventilation_systems"] = ventilation_systems

    # ── lighting systems ───────────────────────────────────────────────────────
    lighting_systems = []
    for i in range(1, 7):
        system = {
            "tech": get(f"lighting_system_tech_z{i}"),
            "ballast": get(f"lighting_system_ballast_z{i}"),
            "lamp_number": get(f"lighting_system_lamp_number_z{i}"),
            "lamp_power": get(f"lighting_system_lamp_power_z{i}"),
            "luminary_number": get(f"lighting_system_luminary_number_z{i}"),
            "name": get(f"lighting_system_name_z{i}"),
            "active_hours": {
                "start": 0,
                "end": get(f"lighting_system_operating_hours_z{i}", 0),
            },
            "count": get(f"lighting_system_similar_zone_number_z{i}"),
            "simultaneity_factor": get(f"lighting_system_simultaneity_factor_z{i}"),
        }
        if has_required_keys(
            system,
            [
                "tech",
                "lamp_number",
                "lamp_power",
                "luminary_number",
                "simultaneity_factor",
            ],
            f"lighting system {i}",
        ):
            lighting_systems.append(system)
    if lighting_systems:
        out["lighting_systems"] = lighting_systems

    # ── hot water systems ──────────────────────────────────────────────────────
    hot_water_systems = []
    water_demand_raw = get("water_demand")
    hot_water_system = {
        "demand": annual_to_consumption(water_demand_raw * 365) if water_demand_raw is not None else None,
        "reference_temperature": get("water_reference_temperature"),
        "supply_temperature": get("water_supply_temperature"),
        "energy_source": get("water_system_energy_source"),
        "efficiency_cop": get("water_system_efficiency_cop"),
        "nominal_capacity": get("water_system_nominal_capacity"),
        "type": get("water_type"),
    }
    if has_required_keys(
        hot_water_system,
        [
            "demand",
            "reference_temperature",
            "supply_temperature",
            "energy_source",
            "efficiency_cop",
            "nominal_capacity",
        ],
        "hot water system",
    ):
        hot_water_systems.append(hot_water_system)
    if hot_water_systems:
        out["hot_water_systems"] = hot_water_systems

    # ── courtyards ─────────────────────────────────────────────────────────────
    courtyards = []
    courtyard = {
        "length": get("courtyard_length"),
        "width": get("courtyard_width"),
        "count": get("courtyard_number"),
    }
    if (
        has_required_keys(courtyard, ["length", "width", "count"], "courtyard")
        and courtyard["count"] > 0
    ):
        courtyards.append(courtyard)
    if courtyards:
        out["courtyards"] = courtyards

    # ── open_courtyards ────────────────────────────────────────────────────────
    open_courtyards: dict = {}
    for o_code, o_name in [
        ("a", "front"),
        ("b", "right"),
        ("c", "back"),
        ("d", "left"),
    ]:
        open_courtyard = {
            "depth": get(f"open_courtyard_depth_{o_code}1"),
            "count": get(f"open_courtyard_number_{o_code}", 0),
        }
        if (
            has_required_keys(open_courtyard, ["depth"], f"open courtyard {o_name}")
            and open_courtyard["count"] > 0
        ):
            open_courtyards[o_name] = open_courtyard
    if open_courtyards:
        out["open_courtyards"] = open_courtyards

    OpenBESSpecificationV2.model_validate(out)

    if not allow_warnings and len(warnings) > 0:
        raise ValueError(
            "Warnings during TOML to JSON conversion:\n" + "\n".join(warnings)
        )

    return out



from typing import List

from pandas import Series
from .. import logging
from pandas import DataFrame

from .base import EnergyUseSimulation, EnergyUseSimulationInitError, SimulationError
from .geometry import BuildingGeometry
from .occupancy import OccupationSimulation
from ..schemas import VentilationSimulationOutput, VentilationSystemResult
from ..types import OpenBESSpecification

logger = logging.getLogger(__name__)


class VentilationSimulationError(SimulationError):
    """Raised when ventilation report generation fails."""


class VentilationSystemSimulation(EnergyUseSimulation):
    """Simulate energy use for a single ventilation system based on building specifications and occupancy patterns."""
    system_number: int
    geometry: BuildingGeometry
    _air_supply_rate_adjusted: float

    def __init__(
        self,
        spec: OpenBESSpecification,
        system_number: int = 1,
        occupancy: OccupationSimulation = None,
        geometry: BuildingGeometry = None,
    ):
        super().__init__(spec=spec)
        self.system_number = system_number
        self.validate_required_inputs(
            required_inputs=[
                "energy_source",
                "airflow",
                "on_time",
                "off_time",
                "rated_input_power",
                "ventilated_area",
            ],
            getter=self._attr,
            context=f"Ventilation system {system_number}",
            error_cls=EnergyUseSimulationInitError,
        )
        try:
            self.geometry = geometry or BuildingGeometry(spec=self.spec)
            self.occupancy = occupancy or OccupationSimulation(spec=self.spec)
        except SimulationError as err:
            raise EnergyUseSimulationInitError(
                f"Failed to initialize VentilationSystemSimulation due to error in dependent simulation: {err}"
            ) from err

    def _attr(self, attr_name: str):
        return self.get_param_or_spec(
            f"ventilation_system{self.system_number}_{attr_name}"
        )

    @property
    def airflow(self) -> float:
        """Rated airflow (m3/h) of the ventilation system."""
        return self._attr("airflow")

    @property
    def air_supply_rate_adjusted(self) -> float:
        """Air supply rate (m3/h/m2) adjusted for system efficiency.
        [Hourly simulation cells IV99, JB99]
        """
        if (
            not hasattr(self, "_air_supply_rate_adjusted")
            or self._air_supply_rate_adjusted is None
        ):
            rated_flow_rate = self.airflow / self._attr("ventilated_area")  # m3/h/m2
            efficiency = self._attr("heat_recovery_efficiency")
            if rated_flow_rate is None or efficiency is None:
                self._air_supply_rate_adjusted = 0.0
            else:
                self._air_supply_rate_adjusted = rated_flow_rate * (1 - efficiency)
        return self._air_supply_rate_adjusted

    @property
    def ventilation_on(self) -> "Series[bool]":
        """Hourly ventilation status (on/off) throughout the year.
        [Hourly simulation columns IR, IX]

        Ventilation only runs between the specified on and off times,
        and only while the building is occupied.
        """
        if "ventilation_on" not in self._hours.columns:
            on_time = self._attr("on_time")
            off_time = self._attr("off_time")
            self._hours["ventilation_on"] = list(
                map(
                    lambda x: on_time <= x <= off_time,
                    self._hours.index.get_level_values("hour").values,
                )
            )
            self._hours["ventilation_on"] = (
                self._hours["ventilation_on"] * self.occupancy.occupied_days
            )
        return self._hours["ventilation_on"]

    @property
    def ventilated_area(self) -> float:
        """Area (m2) served by the ventilation system."""
        return self._attr("ventilated_area")

    @property
    def efficiency(self) -> float:
        """Heat recovery efficiency of the ventilation system."""
        return self._attr("heat_recovery_efficiency")

    @property
    def air_supply_rate(self) -> "Series[float]":
        """Hourly air supply rate (m3/h/m2) throughout the year.
        [Hourly simulation columns IV, JB]
        """
        if "air_supply_rate" not in self._hours.columns:
            area = self.geometry.conditioned_floor_area
            rated_flow_rate = self.air_supply_rate_adjusted * self._attr(
                "ventilated_area"
            )
            if rated_flow_rate is None or area == 0:
                self._hours["air_supply_rate"] = 0.0
            else:
                self._hours["air_supply_rate"] = (
                    rated_flow_rate / area
                ) * self.ventilation_on.astype(float)
        return self._hours["air_supply_rate"]

    @property
    def energy_use(self) -> DataFrame:
        """Ventilation energy use in kWh for each hour of the year for each ENERGY_SOURCE."""
        if self._energy_use[self._attr("energy_source")].hasnans:
            self._energy_use[self._attr("energy_source")] = self.ventilation_on.astype(
                float
            ) * self._attr("rated_input_power")
        return self._energy_use


class VentilationSimulation(EnergyUseSimulation):
    """Aggregate simulation of all ventilation systems in the building, summing their energy use and air supply rates."""
    ventilation_simulations: List[VentilationSystemSimulation]

    def __init__(
        self,
        spec: OpenBESSpecification,
        occupancy: OccupationSimulation = None,
        geometry: BuildingGeometry = None,
    ):
        super().__init__(spec=spec)
        try:
            self.geometry = geometry or BuildingGeometry(spec=self.spec)
            self.occupancy = occupancy or OccupationSimulation(
                spec=self.spec, geometry=self.geometry
            )
        except SimulationError as err:
            raise EnergyUseSimulationInitError(
                f"Failed to initialize VentilationSimulation due to error in dependent simulation: {err}"
            ) from err
        self.ventilation_simulations = []
        while True:
            system_number = len(self.ventilation_simulations) + 1
            try:
                self.ventilation_simulations.append(
                    VentilationSystemSimulation(
                        spec=spec,
                        system_number=system_number,
                        occupancy=self.occupancy,
                        geometry=self.geometry,
                    )
                )
            except EnergyUseSimulationInitError:
                break

    @property
    def air_supply_rate(self) -> "Series[float]":
        """Total hourly air supply rate (m3/h/m2) from all ventilation systems.
        [Hourly simulation column JA]
        """
        if "air_supply_rate" not in self._hours.columns:
            total_air_supply = Series([0.0] * len(self._hours), index=self._hours.index)
            for sim in self.ventilation_simulations:
                total_air_supply += sim.air_supply_rate
            self._hours["air_supply_rate"] = total_air_supply
        return self._hours["air_supply_rate"]

    @property
    def energy_use(self) -> "Series[float]":
        """Ventilation energy use in kWh for each hour of the year for each ENERGY_SOURCES."""
        if len(self.ventilation_simulations) == 0:
            return self._energy_use.fillna(0.0)
        return sum([x.energy_use for x in self.ventilation_simulations])

    @property
    def report(self) -> VentilationSimulationOutput:
        try:
            systems = []
            for vs in self.ventilation_simulations:
                hourly_energy = vs.energy_use.sum(axis="columns")
                ach_hours = (vs.air_supply_rate != 0).sum()
                systems.append(
                    VentilationSystemResult(
                        energy_demand=vs.energy_use.sum().sum() / vs.ventilated_area,
                        peak_load=max(hourly_energy),
                        sfp=max(hourly_energy) / (vs.airflow / 3600),
                        mechanical_ventilation_rate=vs.airflow / vs.ventilated_area,
                        ventilation_rate=vs.airflow / vs.ventilated_area / 3.6,
                        ach=(
                            (vs.air_supply_rate / self.spec.floor_to_ceiling_height).sum()
                            / ach_hours
                            if ach_hours > 0
                            else None
                        ),
                    )
                )
            return VentilationSimulationOutput(ventilation_systems=systems)
        except Exception as exc:
            raise VentilationSimulationError(
                "Failed to generate ventilation report"
            ) from exc



"""Converting OpenBESSpecificationV2 JSON to legacy TOML.

The historical TOML format used by OpenBES only captures a flattened view of
the data model.  The new `openbes.schemas.OpenBESSpecificationV2`
model is richer (allowing arbitrary list lengths, nested structures, etc.), so
these helpers provide a best-effort conversion layer.
"""

from __future__ import annotations

import json
from typing import Any, List, Tuple, Dict

from openbes.schemas import (
    CoolingSystem,
    HeatingSystem,
    VentilationSystem,
)
from openbes.schemas import (
    Consumption,
    OpenBESSpecificationV2,
    ZoneSimultaneity,
)
from openbes.types import meteorological_file_path_to_toml_value


def _len(obj: Any) -> int:
    """Return the length of the object, or 0 if it is None."""
    if obj is None:
        return 0
    return len(obj)


def json_to_toml(
    spec: OpenBESSpecificationV2 | Dict[str, Any] | str, allow_warnings: bool = True
) -> dict:
    """Convert an OpenBESSpecificationV2 or its JSON representation to a TOML mapping.

    Only TOML keys that correspond to fields actually present in the incoming JSON are
    emitted.  Pydantic defaults that were not explicitly set are silently skipped, so
    an empty input ({}) produces an empty output ({}).

    Args:
        spec: The OpenBESSpecificationV2 instance, dictionary, or JSON string to convert.
        allow_warnings: If True, allow lossy conversions with warnings. If False, raise errors on lossy conversions.
    Returns:
        A dictionary representing the TOML mapping (empty-string values are not included).
    Raises:
        ValidationError: If the input does not validate against OpenBESSpecificationV2.
        ValueError: If the conversion is lossy and allow_warnings is False.
    """
    if isinstance(spec, str):
        spec = OpenBESSpecificationV2(**json.loads(spec))
    elif isinstance(spec, dict):
        spec = OpenBESSpecificationV2(**spec)

    toml: dict = {}

    def set_if(key: str, value: Any) -> None:
        """Add *key* to the output only when *value* is not None/empty-string."""
        if value is not None and value != "":
            toml[key] = value

    def annual_consumption(consumption: Consumption) -> float:
        if consumption is None:
            return 0.0
        return sum(dict(consumption).values())

    # ── parameters ─────────────────────────────────────────────────────────────
    p = spec.parameters
    if p is not None:
        pset = p.model_fields_set

        # Each parameter field maps 1-to-1 to a TOML d. key.
        for attr, toml_key in [
            ("altitude", "d.altitude"),
            ("cooling_load_factor", "d.cooling_load_factor"),
            ("density_of_air", "d.density_of_air"),
            ("facade_absorption_coefficient", "d.facade_absorption_coefficient"),
            ("facade_correction_factor", "d.facade_correction_factor"),
            ("facade_emissivity", "d.facade_emissivity"),
            ("floor_correction_factor", "d.floor_correction_factor"),
            ("heat_capacity_correction_factor", "d.heat_capacity_correction_factor"),
            ("heat_capacity_joule", "d.heat_capacity_joule"),
            ("heating_load_factor", "d.heating_load_factor"),
            ("infiltration_correction_factor", "d.infiltration_correction_factor"),
            ("leakage_air_flow_dependent", "d.leakage_air_flow_dependent"),
            ("nia_gba_ratio", "d.nia_gba_ratio"),
            ("pressure_of_air", "d.pressure_of_air"),
            ("roof_absorption_coefficient", "d.roof_absorption_coefficient"),
            ("roof_correction_factor", "d.roof_correction_factor"),
            ("roof_emissivity", "d.roof_emissivity"),
            ("shading_correction_factor", "d.shading_correction_factor"),
            ("specific_heat_of_air", "d.specific_heat_of_air"),
            ("view_factor_to_sky_facade", "d.view_factor_to_sky_facade"),
            ("view_factor_to_sky_roof", "d.view_factor_to_sky_roof"),
            ("window_correction_factor", "d.window_correction_factor"),
        ]:
            if attr in pset:
                set_if(toml_key, getattr(p, attr))

        if "include_appliances" in pset:
            toml["d.appliance_on_off"] = 1 if p.include_appliances else 0
        if "include_lighting" in pset:
            toml["d.lighting_on_off"] = 1 if p.include_lighting else 0
        if "include_occupancy" in pset:
            toml["d.occupancy_on_off"] = 1 if p.include_occupancy else 0

        if "window_angular_correction_factors" in pset and p.window_angular_correction_factors:
            for idx, val in enumerate(p.window_angular_correction_factors):
                set_if(f"d.window_optical_c{idx + 1}", val)

    # ── building ───────────────────────────────────────────────────────────────
    b = spec.building
    if b is not None:
        bset = b.model_fields_set

        for attr, toml_key in [
            ("area", "i.building_area"),
            ("height", "i.building_height"),
            ("length", "i.building_length"),
            ("width", "i.building_width"),
            ("name", "i.building_name"),
            ("type", "i.building_type"),
        ]:
            if attr in bset:
                set_if(toml_key, getattr(b, attr))

        for attr, toml_key in [
            ("floor_to_ceiling_height", "i.floor_to_ceiling_height"),
            ("slab_thickness", "i.slab_thickness"),
            ("orientation_angle", "i.orientation_angle"),
            ("roof_angle", "i.roof_angle"),
            ("solar_external_shading_summer", "i.solar_external_shading_summer"),
            ("solar_external_shading_winter", "i.solar_external_shading_winter"),
            ("terrain_class", "i.terrain_class"),
            ("uvalue_facade", "i.uvalue_facade"),
            ("uvalue_roof", "i.uvalue_roof"),
            ("uvalue_window", "i.uvalue_window"),
            ("uvalue_floor", "i.uvalue_floor"),
            ("window_frame_factor", "i.window_frame_factor"),
            ("window_gvalue", "i.window_gvalue"),
            ("window_height", "i.window_height"),
            ("window_length", "i.window_length"),
        ]:
            if attr in bset:
                set_if(toml_key, getattr(b, attr))

        # thermal_bridge booleans: only emit when explicitly set
        for attr, toml_key in [
            ("thermal_bridge_facade_ground", "i.thermal_bridge_facade_ground"),
            ("thermal_bridge_facade_intermediate", "i.thermal_bridge_facade_intermediate"),
            ("thermal_bridge_facade_roof", "i.thermal_bridge_facade_roof"),
            ("thermal_bridge_shading", "i.thermal_bridge_shading"),
            ("thermal_bridge_window", "i.thermal_bridge_window"),
        ]:
            if attr in bset:
                toml[toml_key] = "Yes" if getattr(b, attr) else "No"

        if "window_counts" in bset and b.window_counts is not None:
            for o_code, o_name in [("a", "front"), ("b", "right"), ("c", "back"), ("d", "left")]:
                counts = getattr(b.window_counts, o_name, None)
                if counts is not None:  # emit even if all-zero; None means side not set
                    floors = ["ground", "first", "second", "third", "fourth"]
                    for f_idx, floor in enumerate(floors):
                        if f_idx < len(counts):
                            # Use toml[] directly so 0 is preserved (set_if skips 0)
                            toml[f"i.window_number_{floor}_{o_code}1"] = counts[f_idx]

    # ── zones ──────────────────────────────────────────────────────────────────
    zone_map = [
        {"default_name": "office", "default_str": "Office area",
         "zone": spec.zones[0] if _len(spec.zones) >= 1 else None},
        {"default_name": "teaching", "default_str": "Teaching",
         "zone": spec.zones[1] if _len(spec.zones) >= 2 else None},
        {"default_name": "canteen", "default_str": "Canteen", "zone": spec.zones[2] if _len(spec.zones) >= 3 else None},
        {"default_name": "common", "default_str": "Common areas",
         "zone": spec.zones[3] if _len(spec.zones) >= 4 else None},
        {"default_name": "other", "default_str": "Other spaces",
         "zone": spec.zones[4] if _len(spec.zones) >= 5 else None},
    ]

    if spec.zones:
        for idx, zm in enumerate(zone_map):
            zone = zm["zone"]
            n = idx + 1
            if zone is None:
                continue
            toml[f"i.zone_name_z{n}"] = zone.name
            zset = zone.model_fields_set
            if "conditioned" in zset:
                toml[f"i.condition_z{n}"] = "Conditioned" if zone.conditioned else "Unconditioned"
            if "active_hours" in zset and zone.active_hours is not None:
                z_name = zm["default_name"]
                # First 3 zones: i. prefix; last 2: d. prefix
                prefix = "d" if idx >= 3 else "i"
                toml[f"{prefix}.occupancy_open_{z_name}"] = zone.active_hours.start
                toml[f"{prefix}.occupancy_close_{z_name}"] = zone.active_hours.end
            if "areas" in zset:
                floors = ["ground", "first", "second", "third", "fourth"]
                for f_idx, floor in enumerate(floors):
                    if f_idx < len(zone.areas):
                        set_if(f"i.{floor}_floor_area_z{n}", zone.areas[f_idx])

    # ── occupation schedule ────────────────────────────────────────────────────
    if spec.occupation_schedule is not None:
        oset = spec.occupation_schedule.model_fields_set
        for k, v in dict(spec.occupation_schedule).items():
            if k in oset and v is not None:
                toml[f"i.schedule_{k.lower()}"] = v

    # ── heat capacity ──────────────────────────────────────────────────────────
    if spec.heat_capacity is not None:
        hc = spec.heat_capacity
        if hc.Am is not None:
            if hc.Cm == 80_000.0:
                toml["i.heat_capacity"] = "Very light"
            elif hc.Cm == 110_000.0:
                toml["i.heat_capacity"] = "Light"
            elif hc.Cm == 165_000.0:
                toml["i.heat_capacity"] = "Medium"
            elif hc.Cm == 260_000.0:
                toml["i.heat_capacity"] = "Heavy"
            elif hc.Cm == 370_000.0:
                toml["i.heat_capacity"] = "Very heavy"
            else:
                toml["i.heat_capacity"] = "Custom Value"
                toml["d.advanced_heat_capacity_am"] = hc.Am

    # ── simple top-level scalars ───────────────────────────────────────────────
    sset = spec.model_fields_set

    for attr, toml_key in [
        ("appliances_load", "i.appliances_load"),
        ("country", "i.country"),
        ("leakage_air_flow", "i.leakage_air_flow"),
        ("leakage_air_flow_independent", "i.leakage_air_flow_independent"),
        ("lighting_control", "i.lighting_control"),
        ("lighting_simultaneity_factor", "i.lighting_simultaneity_factor"),
        ("max_building_occupation", "i.max_building_occupation"),
        ("natural_ventilation_night", "i.natural_ventilation_night"),
        ("typical_occupation", "i.typical_occupation"),
    ]:
        if attr in sset:
            set_if(toml_key, getattr(spec, attr))

    if "holiday" in sset:
        toml["i.holiday"] = "Yes" if spec.holiday else "No"

    if "lighting_active_hours" in sset and spec.lighting_active_hours is not None:
        toml["i.lighting_on_time"] = spec.lighting_active_hours.start
        toml["i.lighting_off_time"] = spec.lighting_active_hours.end

    if "meteorological_file_path" in sset:
        set_if("i.meteorological_file", meteorological_file_path_to_toml_value(
            spec.meteorological_file_path
        ))

    if "setpoint_temperature_day" in sset and spec.setpoint_temperature_day is not None:
        set_if("i.setpoint_summer_day", spec.setpoint_temperature_day.min)
        set_if("i.setpoint_winter_day", spec.setpoint_temperature_day.max)

    if "setpoint_temperature_night" in sset and spec.setpoint_temperature_night is not None:
        set_if("i.setpoint_summer_night", spec.setpoint_temperature_night.min)
        set_if("i.setpoint_winter_night", spec.setpoint_temperature_night.max)

    # ── consumption fields ─────────────────────────────────────────────────────
    for attr, toml_key in [
        ("biomass_consumption", "i.biomass_annual"),
        ("biomass_pellets_consumption", "i.biomass_pellets_annual"),
        ("diesel_consumption", "i.diesel_annual"),
        ("LPG_consumption", "i.LPG_annual"),
    ]:
        if attr in sset and getattr(spec, attr) is not None:
            toml[toml_key] = annual_consumption(getattr(spec, attr))

    if "electricity_consumption" in sset and spec.electricity_consumption is not None:
        for month, v in dict(spec.electricity_consumption).items():
            set_if(f"i.electricity_{month.lower()}", v)

    if "natural_gas_consumption" in sset and spec.natural_gas_consumption is not None:
        for month, v in dict(spec.natural_gas_consumption).items():
            set_if(f"i.gas_{month.lower()}", v)

    if "other_electricity_consumption" in sset and spec.other_electricity_consumption is not None:
        toml["i.other_electricity_usage"] = annual_consumption(spec.other_electricity_consumption) / 12

    if "other_gas_consumption" in sset and spec.other_gas_consumption is not None:
        toml["i.other_gas_usage"] = annual_consumption(spec.other_gas_consumption) / 12

    if "building_standby_electricity_consumption" in sset and spec.building_standby_electricity_consumption is not None:
        toml["i.building_standby_load"] = annual_consumption(spec.building_standby_electricity_consumption) / 12

    if "energy_generated" in sset and spec.energy_generated is not None:
        toml["i.energy_generated"] = annual_consumption(spec.energy_generated)

    if "energy_used" in sset and spec.energy_used is not None:
        toml["i.energy_used"] = annual_consumption(spec.energy_used)

    # ── heating systems ────────────────────────────────────────────────────────
    def get_simultaneity_factors(zs: ZoneSimultaneity) -> List[Tuple[str, Any]]:
        out = []
        for zm in zone_map:
            zone = zm["zone"]
            if zone is not None and zs is not None and zone.name in zs.root:
                out.append((zm["default_name"], zs.root[zone.name]))
        return out

    def heating_system_to_toml(hs: HeatingSystem, n: int) -> None:
        d = "d" if n == 2 else "i"
        set_if(f"{d}.heating_system{n}_energy_source", hs.energy_source)
        set_if(f"{d}.heating_system{n}_efficiency_cop", hs.efficiency_cop)
        set_if(f"{d}.heating_system{n}_nominal_capacity", hs.nominal_capacity)
        # min_demand is always a d. (parameter) key regardless of system number,
        # because from_toml passes d. keys to OpenBESParameters, not OpenBESSpecification.
        set_if(f"d.heating_system{n}_min_demand", hs.min_demand)
        set_if(f"{d}.heating_system{n}_number", hs.count)
        set_if(f"{d}.heating_system{n}_type", hs.type)
        if hs.active_hours is not None:
            toml[f"{d}.heating_system{n}_on_time"] = hs.active_hours.start
            toml[f"{d}.heating_system{n}_off_time"] = hs.active_hours.end
        for zone_name, factor in get_simultaneity_factors(hs.simultaneity):
            toml[f"{d}.heating_system{n}_simultaneity_factor_{zone_name}"] = factor

    if spec.heating_systems:
        for i, hs in enumerate(spec.heating_systems[:2], start=1):
            heating_system_to_toml(hs, i)

    # ── cooling systems ────────────────────────────────────────────────────────
    def cooling_system_to_toml(cs: CoolingSystem, n: int) -> None:
        d = "d" if n == 2 else "i"
        set_if(f"{d}.cooling_system{n}_energy_source", cs.energy_source)
        set_if(f"{d}.cooling_system{n}_energy_efficifiency_ratio", cs.efficiency_ratio)
        set_if(f"{d}.cooling_system{n}_nominal_capacity", cs.nominal_capacity)
        set_if(f"{d}.cooling_system{n}_sensible_nominal_capacity", cs.sensible_nominal_capacity)
        # min_demand is always a d. (parameter) key regardless of system number.
        set_if(f"d.cooling_system{n}_min_demand", cs.min_demand)
        set_if(f"{d}.cooling_system{n}_number", cs.count)
        set_if(f"{d}.cooling_system{n}_type", cs.type)
        if cs.active_hours is not None:
            toml[f"{d}.cooling_system{n}_on_time"] = cs.active_hours.start
            toml[f"{d}.cooling_system{n}_off_time"] = cs.active_hours.end
        for zone_name, factor in get_simultaneity_factors(cs.simultaneity):
            toml[f"{d}.cooling_system{n}_simultaneity_factor_{zone_name}"] = factor

    if spec.cooling_systems:
        for i, cs in enumerate(spec.cooling_systems[:2], start=1):
            cooling_system_to_toml(cs, i)

    # ── ventilation systems ────────────────────────────────────────────────────
    def ventilation_system_to_toml(vs: VentilationSystem, n: int) -> None:
        d = "d" if n == 2 else "i"
        set_if(f"{d}.ventilation_system{n}_energy_source", vs.energy_source)
        set_if(f"{d}.ventilation_system{n}_airflow", vs.airflow)
        set_if(f"{d}.ventilation_system{n}_heat_recovery_efficiency", vs.heat_recovery_efficiency)
        set_if(f"{d}.ventilation_system{n}_rated_input_power", vs.rated_input_power)
        set_if(f"{d}.ventilation_system{n}_type", vs.type)
        set_if(f"{d}.ventilation_system{n}_ventilated_area", vs.ventilated_area)
        if vs.active_hours is not None:
            toml[f"{d}.ventilation_system{n}_on_time"] = vs.active_hours.start
            toml[f"{d}.ventilation_system{n}_off_time"] = vs.active_hours.end

    if spec.ventilation_systems:
        for i, vs in enumerate(spec.ventilation_systems[:2], start=1):
            ventilation_system_to_toml(vs, i)

    # ── lighting systems ───────────────────────────────────────────────────────
    if spec.lighting_systems:
        for i, ls in enumerate(spec.lighting_systems[:6], start=1):
            set_if(f"i.lighting_system_tech_z{i}", ls.tech)
            set_if(f"i.lighting_system_ballast_z{i}", ls.ballast)
            set_if(f"i.lighting_system_lamp_number_z{i}", ls.lamp_number)
            set_if(f"i.lighting_system_lamp_power_z{i}", ls.lamp_power)
            set_if(f"i.lighting_system_luminary_number_z{i}", ls.luminary_number)
            set_if(f"i.lighting_system_name_z{i}", ls.name)
            set_if(f"i.lighting_system_similar_zone_number_z{i}", ls.count)
            set_if(f"i.lighting_system_simultaneity_factor_z{i}", ls.simultaneity_factor)
            if ls.active_hours is not None:
                toml[f"i.lighting_system_operating_hours_z{i}"] = ls.active_hours.end - ls.active_hours.start

    # ── hot water systems ──────────────────────────────────────────────────────
    if spec.hot_water_systems:
        ws = spec.hot_water_systems[0]
        if ws.demand is not None:
            toml["i.water_demand"] = annual_consumption(ws.demand) / 365
        set_if("i.water_reference_temperature", ws.reference_temperature)
        set_if("i.water_supply_temperature", ws.supply_temperature)
        set_if("i.water_system_energy_source", ws.energy_source)
        set_if("i.water_system_efficiency_cop", ws.efficiency_cop)
        set_if("i.water_system_nominal_capacity", ws.nominal_capacity)
        set_if("i.water_system_type", ws.type)

    # ── courtyards ─────────────────────────────────────────────────────────────
    if spec.courtyards:
        if len(spec.courtyards) > 1:
            toml["d.courtyard_length"] = sum(c.length * c.count for c in spec.courtyards)
            toml["d.courtyard_width"] = sum(c.width * c.count for c in spec.courtyards)
            toml["d.courtyard_number"] = 1
        else:
            c = spec.courtyards[0]
            toml["d.courtyard_length"] = c.length
            toml["d.courtyard_width"] = c.width
            set_if("d.courtyard_number", c.count)

    # ── open courtyards ────────────────────────────────────────────────────────
    if spec.open_courtyards is not None:
        for o_code, o_name in [("a", "front"), ("b", "right"), ("c", "back"), ("d", "left")]:
            oc = getattr(spec.open_courtyards, o_name, None)
            if oc is not None:
                toml[f"d.open_courtyard_depth_{o_code}1"] = oc.depth
                toml[f"d.open_courtyard_number_{o_code}"] = oc.count

    return toml
